"""
main
===============

Runs each part of the brainmapper pipeline in turn.
"""

import logging
import os
from datetime import datetime

import brainglobe_space as bgs
import pandas as pd
import tifffile
from brainglobe_utils.cells.cells import MissingCellsError
from brainglobe_utils.general.system import ensure_directory_exists
from brainglobe_utils.image.heatmap import heatmap_from_points
from brainglobe_utils.IO.cells import get_cells, save_cells
from brainglobe_utils.IO.image.load import read_z_stack

BRAINREG_PRE_PROCESSING_ARGS = None


def get_downsampled_space(atlas, downsampled_image_path):
    target_shape = tifffile.imread(downsampled_image_path).shape
    downsampled_space = bgs.AnatomicalSpace(
        atlas.metadata["orientation"],
        shape=target_shape,
        resolution=atlas.resolution,
    )
    return downsampled_space


def cells_exist(points_file):
    try:
        get_cells(points_file, cells_only=True)
        return True
    except MissingCellsError:
        return False


def main():
    from brainreg.core.main import main as register

    from brainglobe_workflows.brainmapper import prep

    start_time = datetime.now()
    args, arg_groups, what_to_run, atlas = prep.prep_brainmapper_general()

    if what_to_run.register:
        # TODO: add register_part_brain option
        logging.info("Registering to atlas")
        args, additional_images_downsample = prep.prep_registration(args)
        register(
            args.atlas,
            args.orientation,
            args.target_brain_path,
            args.brainreg_paths,
            args.voxel_sizes,
            arg_groups["NiftyReg registration backend options"],
            BRAINREG_PRE_PROCESSING_ARGS,
            sort_input_file=args.sort_input_file,
            n_free_cpus=args.n_free_cpus,
            additional_images_downsample=additional_images_downsample,
            backend=args.backend,
            debug=args.debug,
        )

    else:
        logging.info("Skipping registration")

    if len(args.signal_planes_paths) > 1:
        base_directory = args.output_dir

        for idx, signal_paths in enumerate(args.signal_planes_paths):
            channel = args.signal_ch_ids[idx]
            logging.info("Processing channel: " + str(channel))
            channel_directory = os.path.join(
                base_directory, "channel_" + str(channel)
            )
            if not os.path.exists(channel_directory):
                os.makedirs(channel_directory)

            # prep signal channel specific args
            args.signal_planes_paths[0] = signal_paths
            # TODO: don't overwrite args.output_dir - use Paths instead
            args.output_dir = channel_directory
            args.signal_channel = channel
            # Run for each channel
            run_all(args, what_to_run, atlas)

    else:
        args.signal_channel = args.signal_ch_ids[0]
        run_all(args, what_to_run, atlas)
    logging.info(
        "Finished. Total time taken: {}".format(datetime.now() - start_time)
    )


def run_all(args, what_to_run, atlas):
    from cellfinder.core.classify import classify
    from cellfinder.core.detect import detect
    from cellfinder.core.tools import prep

    from brainglobe_workflows.brainmapper import analyse
    from brainglobe_workflows.brainmapper.prep import (
        prep_candidate_detection,
        prep_channel_specific_general,
    )

    points = None
    signal_array = None
    args, what_to_run = prep_channel_specific_general(args, what_to_run)

    if what_to_run.detect:
        logging.info("Detecting cell candidates")
        args = prep_candidate_detection(args)
        signal_array = read_z_stack(
            args.signal_planes_paths[args.signal_channel]
        )

        points = detect.main(
            signal_array=signal_array,
            start_plane=args.start_plane,
            end_plane=args.end_plane,
            voxel_sizes=args.voxel_sizes,
            soma_diameter=args.soma_diameter,
            max_cluster_size=args.max_cluster_size,
            ball_xy_size=args.ball_xy_size,
            ball_z_size=args.ball_z_size,
            ball_overlap_fraction=args.ball_overlap_fraction,
            soma_spread_factor=args.soma_spread_factor,
            n_free_cpus=args.n_free_cpus,
            log_sigma_size=args.log_sigma_size,
            n_sds_above_mean_thresh=args.n_sds_above_mean_thresh,
            n_sds_above_mean_tiled_thresh=args.n_sds_above_mean_tiled_thresh,
            tiled_thresh_tile_size=args.tiled_thresh_tile_size,
            save_planes=args.save_planes,
            plane_directory=args.plane_directory,
            batch_size=args.detection_batch_size,
            torch_device=args.torch_device,
            pin_memory=args.pin_memory,
        )
        ensure_directory_exists(args.paths.points_directory)

        save_cells(
            points,
            args.paths.detected_points,
            save_csv=args.save_csv,
            artifact_keep=args.artifact_keep,
        )

    else:
        logging.info("Skipping cell detection")
        points = get_cells(args.paths.detected_points)

    if what_to_run.classify:
        model_weights = prep.prep_model_weights(
            args.model_weights,
            args.install_path,
            args.model,
        )
        if what_to_run.classify:
            if points is None:
                points = get_cells(args.paths.detected_points)
            if signal_array is None:
                signal_array = read_z_stack(
                    args.signal_planes_paths[args.signal_channel]
                )
            logging.info("Running cell classification")
            background_array = read_z_stack(args.background_planes_path[0])

            points = classify.main(
                points=points,
                signal_array=signal_array,
                background_array=background_array,
                n_free_cpus=args.n_free_cpus,
                voxel_sizes=args.voxel_sizes,
                network_voxel_sizes=args.network_voxel_sizes,
                batch_size=args.classification_batch_size,
                cube_height=args.cube_height,
                cube_width=args.cube_width,
                cube_depth=args.cube_depth,
                trained_model=args.trained_model,
                model_weights=model_weights,
                network_depth=args.network_depth,
                pin_memory=args.pin_memory,
            )
            save_cells(
                points,
                args.paths.classified_points,
                save_csv=args.save_csv,
            )

            what_to_run.cells_exist = cells_exist(args.paths.classified_points)

        else:
            logging.info("No cells were detected, skipping classification.")

    else:
        logging.info("Skipping cell classification")

    what_to_run.update_if_cells_required()

    if what_to_run.analyse or what_to_run.figures:
        downsampled_space = get_downsampled_space(
            atlas, args.brainreg_paths.boundaries_file_path
        )

    if what_to_run.analyse:
        points = get_cells(args.paths.classified_points, cells_only=True)
        if len(points) == 0:
            logging.info("No cells detected, skipping cell position analysis")
        else:
            logging.info("Analysing cell positions")
            analyse.run(args, points, atlas, downsampled_space)
    else:
        logging.info("Skipping cell position analysis")

    if what_to_run.figures:
        points = get_cells(args.paths.classified_points, cells_only=True)
        if len(points) == 0:
            logging.info("No cells detected, skipping")
        else:
            logging.info("Generating heatmap")

            if args.mask_figures:
                mask_image = tifffile.imread(
                    args.brainreg_paths.registered_atlas
                )
            else:
                mask_image = None

            downsampled_points = pd.read_hdf(
                args.paths.downsampled_points
            ).values

            heatmap_from_points(
                downsampled_points,
                atlas.resolution[0],  # assumes isotropic atlas
                downsampled_space.shape,
                output_filename=args.paths.heatmap,
                smoothing=args.heatmap_smooth,
                mask_image=mask_image,
            )
    else:
        logging.info("Skipping figure generation")


if __name__ == "__main__":
    main()



from importlib.metadata import PackageNotFoundError, metadata

try:
    __version__ = metadata("brainglobe_workflows")["version"]
    __author__ = metadata("brainglobe_workflows")["author-email"]
    __license__ = metadata("brainglobe_workflows")["license"]
except PackageNotFoundError:
    # Package not installed
    pass

del metadata



import argparse
import logging
import sys
from pathlib import Path
from typing import List

DEFAULT_JSON_CONFIGS_PATH = Path(__file__).resolve().parent / "configs"

DEFAULT_JSON_CONFIG_PATH_CELLFINDER = (
    DEFAULT_JSON_CONFIGS_PATH / "cellfinder.json"
)


def setup_logger() -> logging.Logger:
    """Setup a logger for workflow runs

    The logger's level is set to DEBUG, and it
    is linked to a handler that writes to the
    console. This utility function helps run
    workflows, and test their logs, in a
    consistent way.

    Returns
    -------
    logging.Logger
        a logger object configured for workflow runs
    """
    # define handler that writes to stdout
    console_handler = logging.StreamHandler(sys.stdout)
    console_format = logging.Formatter("%(name)s %(levelname)s: %(message)s")
    console_handler.setFormatter(console_format)
    console_handler.set_name("console_handler")

    # define logger and link to handler
    logger = logging.getLogger(
        __name__
    )  # if imported as a module, the logger is named after the module
    logger.setLevel(logging.DEBUG)
    logger.addHandler(console_handler)
    return logger


def config_parser(
    argv: List[str],
    default_config: str,
) -> argparse.Namespace:
    """Define argument parser for a workflow script.

    The only CLI argument defined in the parser is
    the input config file. The list of input arguments
    `argv` can be an empty list.

    Both the list of input arguments and the default config to use if
    no config is specified must be passed as an input to this
    function.

    Parameters
    ----------
    argv_ : List[str]
        _description_
    default_config : str
        _description_

    Returns
    -------
    args : argparse.Namespace
        command line input arguments parsed
    """

    # initialise argument parser
    parser = argparse.ArgumentParser(
        description=(
            """
            To launch the cellfinder workflow with default parameters, run:
                `cellfinder-workflow`.
            The default parameters are those specifed in brainglobe_workflows/
            cellfinder/configs/cellfinder.json.


            To launch the cellfinder workflow with a specific set of input
            parameters, run:
                `cellfinder-workflow --config path/to/config.json`,
            where `path/to/input/config.json` is the json file with the
            desired parameters.
            """
        ),
        formatter_class=argparse.RawTextHelpFormatter,
    )

    # add arguments
    parser.add_argument(
        "-c",
        "--config",
        default=default_config,
        type=str,
        metavar="CONFIG",  # a name for usage messages
        help="",
    )

    # build parser object
    args = parser.parse_args(argv)

    return args



"""
prep
==================
Functions to prepare files and directories needed for other functions
"""

import json
import logging
import os
from argparse import Namespace
from pathlib import PurePath

from brainglobe_atlasapi import BrainGlobeAtlas
from brainglobe_utils.general.exceptions import CommandLineInputError
from brainglobe_utils.general.list import check_unique_list, common_member
from brainglobe_utils.general.system import (
    catch_input_file_error,
    ensure_directory_exists,
)
from brainreg.core.paths import Paths as BrainRegPaths
from fancylog import fancylog

import brainglobe_workflows as package_for_log
import brainglobe_workflows.brainmapper.parser as parser
from brainglobe_workflows.brainmapper.parser import (
    brainmapper_parser,
)


def get_arg_groups(args, parser):
    arg_groups = {}
    for group in parser._action_groups:
        group_dict = {
            a.dest: getattr(args, a.dest, None) for a in group._group_actions
        }
        arg_groups[group.title] = Namespace(**group_dict)

    return arg_groups


def check_input_arg_existance(args):
    """
    Does a simple check to ensure that input files/paths exist. Prevents a typo
    from causing brainmapper to only run partway. Doesn't check for validity
    etc, just existence.
    :param args: brainmapper input arguments
    """
    check_list = [args.background_planes_path[0]]
    check_list = check_list + args.signal_planes_paths

    for path in check_list:
        if path is not None:
            catch_input_file_error(path)


class Paths:
    """
    A single class to hold all file paths that brainmapper may need. Any paths
    prefixed with "tmp__" refer to internal intermediate steps, and will be
    deleted if "--debug" is not used.
    """

    def __init__(self, output_dir):
        self.output_dir = output_dir
        self.registration_output_folder = os.path.join(
            self.output_dir, "registration"
        )
        self.metadata_path = os.path.join(self.output_dir, "brainmapper.json")
        self.registration_metadata_path = os.path.join(
            self.registration_output_folder, "brainreg.json"
        )

    def make_channel_specific_paths(self):
        self.points_directory = os.path.join(self.output_dir, "points")
        self.detected_points = os.path.join(self.points_directory, "cells.xml")
        self.classified_points = os.path.join(
            self.points_directory, "cell_classification.xml"
        )
        self.downsampled_points = os.path.join(
            self.points_directory, "downsampled.points"
        )
        self.atlas_points = os.path.join(self.points_directory, "atlas.points")
        self.brainrender_points = os.path.join(
            self.points_directory, "points.npy"
        )

        self.tmp__cubes_output_dir = os.path.join(self.output_dir, "cubes")

        self.figures_directory = os.path.join(self.output_dir, "figures")
        self.heatmap = os.path.join(self.figures_directory, "heatmap.tiff")

        self.analysis_directory = os.path.join(self.output_dir, "analysis")
        self.summary_csv = os.path.join(self.analysis_directory, "summary.csv")
        self.all_points_csv = os.path.join(
            self.analysis_directory, "all_points.csv"
        )


def serialise(obj):
    if isinstance(obj, PurePath):
        return str(obj)
    else:
        return obj.__dict__


def log_metadata(file_path, args):
    with open(file_path, "w") as f:
        json.dump(args, f, default=serialise)


def prep_brainmapper_general():
    args = parser.brainmapper_parser().parse_args()
    arg_groups = get_arg_groups(args, brainmapper_parser())

    check_input_arg_existance(args)

    if not os.path.exists(args.output_dir):
        os.makedirs(args.output_dir)

    args.paths = Paths(args.output_dir)

    fancylog.start_logging(
        args.output_dir,
        package=package_for_log,
        variables=[args, args.paths],
        verbose=args.debug,
        log_header="BRAINMAPPER LOG",
        multiprocessing_aware=True,
    )

    log_metadata(args.paths.metadata_path, args)

    what_to_run = CalcWhatToRun(args)
    args.signal_ch_ids, args.background_ch_id = check_and_return_ch_ids(
        args.signal_ch_ids, args.background_ch_id, args.signal_planes_paths
    )
    args.brainreg_paths = BrainRegPaths(args.paths.registration_output_folder)
    atlas = BrainGlobeAtlas(args.atlas)
    return args, arg_groups, what_to_run, atlas


def check_and_return_ch_ids(signal_ids, background_id, signal_channel_list):
    """
    If channel IDs are provided (via CLI), then they are checked for
    suitability (i.e. they are unique).
    If channel IDs are not given, then unique IDs will be generated

    Assumes only 1 background channel (for registration)

    :param list signal_ids: List of ID ints, or None.
    :param int background_id: ID int, or None.
    :param list signal_channel_list: Input files
    :return: list of signal channel IDs, and background channel ID
    """

    # TODO: make this more general, so it gives informative error messages
    #  when called from brainmapper.extract.cli

    num_signal_channels = len(signal_channel_list)

    if signal_ids is None:
        if background_id is None:
            signal_ids = list(range(num_signal_channels))
            background_id = num_signal_channels  # max(signal_ids) + 1
            pass
        else:
            signal_start = background_id + 1
            signal_end = signal_start + num_signal_channels
            signal_ids = list(range(signal_start, signal_end))
            pass
    else:
        check_correct_number_signal_channels(signal_ids, num_signal_channels)
        check_signal_overlap(signal_ids)
        if background_id is None:
            background_id = max(signal_ids) + 1
            pass
        else:
            # N.B. check_id_overlap needs bg_ch to be iterable (e.g. a list)
            check_sig_bg_id_overlap(signal_ids, [background_id])
    return signal_ids, background_id


def check_correct_number_signal_channels(signal_ids, num_signal_channels):
    num_signal_ids = len(signal_ids)
    if num_signal_channels != num_signal_ids:
        logging.error(
            "Number of signal channel IDs ({}) does not match the number of "
            "signal channels ({})".format(num_signal_ids, num_signal_channels)
        )

        raise CommandLineInputError(
            "Number of signal channel IDs ({}) does not match the number of "
            "signal channels ({})".format(num_signal_ids, num_signal_channels)
        )


def check_signal_overlap(signal_ids):
    signal_unique, repeated_signal_ch = check_unique_list(signal_ids)
    if not signal_unique:
        logging.error(
            "Non-unique values for '--signal-channel-ids' ({}) given. ID:"
            " {} given more than once.".format(signal_ids, repeated_signal_ch)
        )
        raise CommandLineInputError(
            "Non-unique values for '--signal-channel-ids' ({}) given. ID:"
            " {} given more than once.".format(signal_ids, repeated_signal_ch)
        )


def check_sig_bg_id_overlap(signal_ids, background_id):
    common_id_result, intersection = common_member(signal_ids, background_id)

    if common_id_result:
        logging.error(
            "Non-unique values for '--signal-channel-ids' ({}) and for "
            "'--background-channel-id' ({}) given. ID: {} was given in"
            " both.".format(signal_ids, background_id, intersection)
        )
        raise CommandLineInputError(
            "Non-unique values for '--signal-channel-ids' ({}) and for "
            "'--background-channel-id' ({}) given. ID: {} was given in"
            " both.".format(signal_ids, background_id, intersection)
        )


class CalcWhatToRun:
    """
    Class to (hopefully) simplify what should and shouldn't be run
    """

    def __init__(self, args):
        self.detect = True
        self.classify = True
        self.register = True
        self.analyse = True
        self.figures = True
        self.candidates_exist = True
        self.cells_exist = True

        self.atlas_image = os.path.join(
            args.paths.registration_output_folder, "registered_atlas.tiff"
        )
        # order is important
        self.cli_options(args)
        self.existence()

    def update(self, args):
        self.cli_options(args)
        self.existence()
        self.channel_specific_update(args)

    def cli_options(self, args):
        self.detect = not args.no_detection
        self.classify = not args.no_classification
        self.register = not args.no_register
        self.analyse = not args.no_analyse
        self.figures = not args.no_figures

    def existence(self):
        if os.path.exists(self.atlas_image):
            logging.warning(
                "Registered atlas exists, assuming " "already run. Skipping."
            )
            self.register = False

    def channel_specific_update(self, args):
        if os.path.exists(args.paths.detected_points):
            logging.warning(
                "Initial detection file exists (cells.xml), "
                "assuming already run. Skipping."
            )
            self.detect = False

        if os.path.exists(args.paths.classified_points):
            logging.warning(
                "Cell classification file "
                "(cell_classification.xml)"
                " exists, assuming already run. Skipping cell detection, "
                "and classification"
            )
            self.classify = False
            self.detect = False

        if not os.path.exists(self.atlas_image):
            self.analyse = False
            self.figures = False

        if os.path.exists(args.paths.summary_csv):
            self.analyse = False

        if os.path.exists(args.paths.heatmap):
            self.figures = False

    def update_if_candidates_required(self):
        if not self.candidates_exist:
            self.classify = False

    def update_if_cells_required(self):
        if not self.cells_exist:
            self.analyse = False
            self.figures = False


def prep_registration(args):
    args.target_brain_path = args.background_planes_path[0]
    logging.debug("Making registration directory")
    ensure_directory_exists(args.paths.registration_output_folder)
    log_metadata(args.paths.registration_metadata_path, args)
    additional_images_downsample = {}
    for idx, images in enumerate(args.signal_planes_paths):
        channel = args.signal_ch_ids[idx]
        additional_images_downsample[f"channel_{channel}"] = images
    return args, additional_images_downsample


def prep_channel_specific_general(args, what_to_run):
    args.paths.output_dir = args.output_dir
    args.paths.make_channel_specific_paths()
    what_to_run.update(args)
    return args, what_to_run


def prep_candidate_detection(args):
    if args.save_planes:
        args.plane_directory = os.path.join(
            args.output_dir, "processed_planes"
        )
        ensure_directory_exists(args.plane_directory)
    else:
        args.plane_directory = None  # FIXME: remove this fudge

    return args



import logging

import numpy as np
from brainglobe_utils.brainmapper.analysis import (
    summarise_points_by_atlas_region,
)
from brainglobe_utils.brainmapper.export import export_points_to_brainrender
from brainglobe_utils.brainreg.transform import transform_points_to_atlas_space


def run(args, cells, atlas, downsampled_space):
    deformation_field_paths = [
        args.brainreg_paths.deformation_field_0,
        args.brainreg_paths.deformation_field_1,
        args.brainreg_paths.deformation_field_2,
    ]

    cell_list = []
    for cell in cells:
        cell_list.append([cell.z, cell.y, cell.x])
    cells = np.array(cell_list)

    logging.info("Transforming points to atlas space")
    transformed_cells, points_out_of_bounds = transform_points_to_atlas_space(
        cells,
        args.signal_planes_paths[0],
        args.orientation,
        args.voxel_sizes,
        downsampled_space,
        atlas,
        deformation_field_paths,
        downsampled_points_path=args.paths.downsampled_points,
        atlas_points_path=args.paths.atlas_points,
    )
    logging.warning(
        f"{len(points_out_of_bounds)} points ignored due to falling outside "
        f"of atlas. This may be due to inaccuracies with "
        f"cell detection or registration. Please inspect the results."
    )

    logging.info("Summarising cell positions")
    summarise_points_by_atlas_region(
        cells,
        transformed_cells,
        atlas,
        args.brainreg_paths.volume_csv_path,
        args.paths.all_points_csv,
        args.paths.summary_csv,
    )
    logging.info("Exporting data to brainrender")
    export_points_to_brainrender(
        transformed_cells, atlas.resolution[0], args.paths.brainrender_points
    )



"""This script reproduces the most common cellfinder workflow

It receives as an (optional) command line input the path to a configuration
json file, that holds the values of the required parameters for the workflow.

    python brainglobe_workflows/cellfinder.py --config path/to/config.json

If no input json file is passed as a configuration, the default
configuration defined at brainglobe_workflows/cellfinder/default_config.json
is used.

    python brainglobe_workflows/cellfinder.py

"""

import datetime
import json
import logging
import os
import sys
from dataclasses import dataclass
from pathlib import Path
from typing import Optional, Union

import pooch
from brainglobe_utils.IO.cells import save_cells
from brainglobe_utils.IO.image.load import read_with_dask
from cellfinder.core.main import main as cellfinder_run
from cellfinder.core.train.train_yaml import depth_type

from brainglobe_workflows.utils import (
    DEFAULT_JSON_CONFIG_PATH_CELLFINDER,
    config_parser,
    setup_logger,
)
from brainglobe_workflows.utils import __name__ as LOGGER_NAME

Pathlike = Union[str, os.PathLike]


@dataclass
class CellfinderConfig:
    """Define the parameters for the cellfinder workflow.

    There are three types of fields:
    - required attributes: must be provided, they do not have a default value;
    - optional attributes: they have a default value if not specified;
    - internal attributes: their names start with _, indicating these are
      private. Any functionality to update them should be a class method.
    """

    # Required parameters
    voxel_sizes: tuple[float, float, float]
    start_plane: int
    end_plane: int
    trained_model: Optional[os.PathLike]
    model_weights: Optional[os.PathLike]
    model: str
    classification_batch_size: int
    n_free_cpus: int
    network_voxel_sizes: tuple[float, float, float]
    soma_diameter: float
    ball_xy_size: float
    ball_z_size: float
    ball_overlap_fraction: float
    log_sigma_size: float
    n_sds_above_mean_thresh: float
    n_sds_above_mean_tiled_thresh: float
    tiled_thresh_tile_size: float | None
    soma_spread_factor: float
    max_cluster_size: float
    cube_width: int
    cube_height: int
    cube_depth: int
    network_depth: depth_type
    detection_batch_size: Optional[int]
    torch_device: Optional[str]
    pin_memory: bool

    # Optional parameters

    # install path: default path for downloaded and output data
    _install_path: Pathlike = (
        Path.home() / ".brainglobe" / "workflows" / "cellfinder"
    )

    # input data paths
    # Note: if not specified, the signal and background data
    # are assumed to be under "signal" and "background"
    # dirs under _install_path/cellfinder_test_data/
    # (see __post_init__ method)
    input_data_dir: Optional[Pathlike] = None
    signal_subdir: Pathlike = "signal"
    background_subdir: Pathlike = "background"

    # output data paths
    # Note: if output_parent_dir is not specified,
    # it is assumed to be under _install_path
    # (see __post_init__ method)
    output_dir_basename: str = "cellfinder_output_"
    detected_cells_filename: str = "detected_cells.xml"
    output_parent_dir: Optional[Pathlike] = None

    # source of data to download
    # if not specified in JSON, it is set to None
    data_url: Optional[str] = None
    data_hash: Optional[str] = None

    # Internal parameters
    # even though these are optional we don't expect users to
    # change them
    _signal_dir_path: Optional[Pathlike] = None
    _background_dir_path: Optional[Pathlike] = None
    _list_signal_files: Optional[list] = None
    _list_background_files: Optional[list] = None
    _detected_cells_path: Pathlike = ""
    _output_path: Pathlike = ""

    def __post_init__(self: "CellfinderConfig"):
        """Executed after __init__ function.

        We use this method to define attributes of the data class
        as a function of other attributes.
        See https://peps.python.org/pep-0557/#post-init-processing

        The attributes added are input and output data paths

        Parameters
        ----------
        self : CellfinderConfig
            a CellfinderConfig instance
        """

        # Add input data paths to config
        self.add_input_paths()

        # Add output paths to config
        self.add_output_paths()

    def add_output_paths(self):
        """Adds output paths to the config

        Specifically, it adds:
        - output_parent_dir: set to a a timestamped output directory if not
          set in __init__();
        - _detected_cells_path: path to the output file

        Parameters
        ----------
        config : CellfinderConfig
            a cellfinder config
        """

        # Fill in output directory if not specified
        if self.output_parent_dir is None:
            self.output_parent_dir = Path(self._install_path)

        # Add to config the path to timestamped output directory
        timestamp = datetime.datetime.now()
        timestamp_formatted = timestamp.strftime("%Y%m%d_%H%M%S")
        self._output_path = Path(self.output_parent_dir) / (
            str(self.output_dir_basename) + timestamp_formatted
        )
        self._output_path.mkdir(
            parents=True,  # create any missing parents
            exist_ok=True,  # ignore FileExistsError exceptions
        )

        # Add to config the path to the output file
        self._detected_cells_path = (
            self._output_path / self.detected_cells_filename
        )

    def add_input_paths(self):
        """Adds input data paths to the config.

        Specifically, it adds:
        - input_data_dir: set to a default value if not set in __init__();
        - _signal_dir_path: full path to the directory with the signal files
        - _background_dir_path: full path to the directory with the
          background files.
        - _list_signal_files: list of signal files
        - _list_background_files: list of background files

        Parameters
        ----------
        config : CellfinderConfig
            a cellfinder config with input data files to be validated

        Notes
        -----
        The signal and background files are first searched locally at the
        given location. If not found, we attempt to download them from GIN
        and place them at the specified location (input_data_dir).

        - If both parent data directories (signal and background) exist
        locally, the lists of signal and background files are added to
        the config.
        - If exactly one of the parent data directories is missing, an error
        message is logged.
        - If neither of them exist, the data is retrieved from the provided GIN
        repository. If no URL or hash to GIN is provided, an error is thrown.

        """
        # Fetch logger
        logger = logging.getLogger(LOGGER_NAME)

        # Fill in input data directory if not specified
        if self.input_data_dir is None:
            self.input_data_dir = (
                Path(self._install_path) / "cellfinder_test_data"
            )

        # Fill in signal and background paths derived from 'input_data_dir'
        self._signal_dir_path = self.input_data_dir / Path(self.signal_subdir)
        self._background_dir_path = self.input_data_dir / Path(
            self.background_subdir
        )

        # Check if input data directories (signal and background) exist
        # locally.
        # If both directories exist, get list of signal and background files
        if (
            Path(self._signal_dir_path).exists()
            and Path(self._background_dir_path).exists()
        ):
            logger.info("Fetching input data from the local directories")

            self._list_signal_files = [
                f
                for f in Path(self._signal_dir_path).resolve().iterdir()
                if f.is_file()
            ]
            self._list_background_files = [
                f
                for f in Path(self._background_dir_path).resolve().iterdir()
                if f.is_file()
            ]

        # If exactly one of the input data directories is missing, print error
        elif (
            Path(self._signal_dir_path).resolve().exists()
            or Path(self._background_dir_path).resolve().exists()
        ):
            if not Path(self._signal_dir_path).resolve().exists():
                logger.error(
                    f"The directory {self._signal_dir_path} does not exist",
                )
            else:
                logger.error(
                    f"The directory {self._background_dir_path} "
                    "does not exist",
                )

        # If neither of the input data directories exist,
        # retrieve data from GIN repository and add list of files to config
        else:
            # Check if GIN URL and hash are defined (log error otherwise)
            if self.data_url and self.data_hash:
                # get list of files in GIN archive with pooch.retrieve
                list_files_archive = pooch.retrieve(
                    url=self.data_url,
                    known_hash=self.data_hash,
                    path=Path(
                        self.input_data_dir
                    ).parent,  # zip will be downloaded here
                    progressbar=True,
                    processor=pooch.Unzip(
                        extract_dir=Path(self.input_data_dir).stem,
                        # files are unpacked here, a dir
                        # *relative* to the path set in 'path'
                    ),
                )
                logger.info(
                    "Fetching input data from the provided GIN repository"
                )

                # Check signal and background parent directories exist now
                assert Path(self._signal_dir_path).resolve().exists()
                assert Path(self._background_dir_path).resolve().exists()

                # Add signal files to config
                self._list_signal_files = [
                    f
                    for f in list_files_archive
                    if f.startswith(
                        str(Path(self._signal_dir_path).resolve()),
                    )
                ]

                # Add background files to config
                self._list_background_files = [
                    f
                    for f in list_files_archive
                    if f.startswith(
                        str(Path(self._background_dir_path).resolve()),
                    )
                ]
            # If one of URL/hash to GIN repo not defined, throw an error
            else:
                logger.error(
                    "Input data not found locally, and URL/hash to "
                    "GIN repository not provided",
                )


def read_cellfinder_config(
    input_config_path: str, log_on: bool = False
) -> CellfinderConfig:
    """Instantiate a CellfinderConfig from the input json file.

    Assumes config is json serializable.

    Parameters
    ----------
    input_config_path : Path
        Absolute path to a cellfinder config file
    log_on : bool, optional
        whether to log the info messages from reading the config
        to the logger, by default False

    Returns
    -------
    CellfinderConfig:
        The cellfinder config object, populated with data from the input
    """

    # read input config
    with open(input_config_path) as cfg:
        config_dict = json.load(cfg)
    config = CellfinderConfig(**config_dict)

    # print config's origin to log if required
    if log_on:
        logger = logging.getLogger(LOGGER_NAME)
        logger.info(f"Input config read from {input_config_path}")
        if input_config_path == DEFAULT_JSON_CONFIG_PATH_CELLFINDER:
            logger.info("Using default config file")

    return config


def setup(input_config_path: str) -> CellfinderConfig:
    # setup logger
    _ = setup_logger()

    # read config
    cfg = read_cellfinder_config(input_config_path)

    return cfg


def run_workflow_from_cellfinder_run(cfg: CellfinderConfig):
    """Run workflow based on the cellfinder.core.main.main()
    function.

    The steps are:
    1. Read the input signal and background data as two separate
       Dask arrays.
    2. Run the main cellfinder pipeline on the input Dask arrays,
       with the parameters defined in the input configuration (cfg).
    3. Save the detected cells as an xml file to the location specified in
       the input configuration (cfg).

    Parameters
    ----------
    cfg : CellfinderConfig
        a class with the required setup methods and parameters for
        the cellfinder workflow
    """
    # Read input data as Dask arrays
    signal_array = read_with_dask(str(cfg._signal_dir_path))
    background_array = read_with_dask(str(cfg._background_dir_path))

    # Run main analysis using `cellfinder_run`
    detected_cells = cellfinder_run(
        signal_array=signal_array,
        background_array=background_array,
        voxel_sizes=cfg.voxel_sizes,
        start_plane=cfg.start_plane,
        end_plane=cfg.end_plane,
        trained_model=cfg.trained_model,
        model_weights=cfg.model_weights,
        model=cfg.model,
        classification_batch_size=cfg.classification_batch_size,
        n_free_cpus=cfg.n_free_cpus,
        network_voxel_sizes=cfg.network_voxel_sizes,
        soma_diameter=cfg.soma_diameter,
        ball_xy_size=cfg.ball_xy_size,
        ball_z_size=cfg.ball_z_size,
        ball_overlap_fraction=cfg.ball_overlap_fraction,
        log_sigma_size=cfg.log_sigma_size,
        n_sds_above_mean_thresh=cfg.n_sds_above_mean_thresh,
        n_sds_above_mean_tiled_thresh=cfg.n_sds_above_mean_tiled_thresh,
        tiled_thresh_tile_size=cfg.tiled_thresh_tile_size,
        soma_spread_factor=cfg.soma_spread_factor,
        max_cluster_size=cfg.max_cluster_size,
        cube_width=cfg.cube_width,
        cube_height=cfg.cube_height,
        cube_depth=cfg.cube_depth,
        network_depth=cfg.network_depth,
        detection_batch_size=cfg.detection_batch_size,
        torch_device=cfg.torch_device,
        pin_memory=cfg.pin_memory,
    )

    # Save results to xml file
    save_cells(
        detected_cells,
        cfg._detected_cells_path,
    )


def main(
    input_config: str = str(DEFAULT_JSON_CONFIG_PATH_CELLFINDER),
) -> CellfinderConfig:
    """Setup and run cellfinder workflow.

    This function runs the setup steps required
    to run the cellfinder workflow, and the
    workflow itself. Note that only the workflow
    will be benchmarked.

    Parameters
    ----------
    input_config : str, optional
        Absolute path to input config file,
        by default str(DEFAULT_JSON_CONFIG_PATH_CELLFINDER)

    Returns
    -------
    cfg : CellfinderConfig
        a class with the required setup methods and parameters for
        the cellfinder workflow
    """
    # run setup
    cfg = setup(input_config)

    # run workflow
    run_workflow_from_cellfinder_run(cfg)  # only this will be benchmarked

    return cfg


if __name__ == "__main__":
    # parse CLI arguments
    args = config_parser(
        sys.argv[1:],
        str(DEFAULT_JSON_CONFIG_PATH_CELLFINDER),
    )

    # run setup and workflow
    _ = main(args.config)



"""
parser
========
All the various arguments that may be needed by the various submodules.
Defined together so they can also be called by any other entry points.
"""

from argparse import (
    ArgumentDefaultsHelpFormatter,
    ArgumentParser,
    ArgumentTypeError,
)
from functools import partial

from brainglobe_utils.general.numerical import (
    check_positive_float,
    check_positive_int,
)
from brainglobe_utils.general.string import check_str
from brainreg.core.cli import atlas_parse, geometry_parser, niftyreg_parse
from brainreg.core.cli import backend_parse as brainreg_backend_parse
from cellfinder.core.download.cli import download_parser
from cellfinder.core.tools.source_files import user_specific_configuration_path

from brainglobe_workflows import __version__

# TODO: Gradually move all paths as strings to Path objects

models = {
    "18": "18-layer",
    "34": "34-layer",
    "50": "50-layer",
    "101": "101-layer",
    "152": "152-layer",
}


def valid_model_depth(depth):
    """
    Ensures a correct existing_model is chosen
    :param value: Input value
    :param models: Dict of allowed models
    :return: Input value, if it corresponds to a valid existing_model
    """

    if depth in models.keys():
        return depth
    else:
        raise ArgumentTypeError(
            f"Model depth: {depth} is not valid. Please "
            f"choose one of: {list(models.keys())}"
        )


def brainmapper_parser():
    parser = ArgumentParser(formatter_class=ArgumentDefaultsHelpFormatter)
    parser = main_parse(parser)
    parser = config_parse(parser)
    parser = pixel_parser(parser)
    parser = run_parse(parser)
    parser = io_parse(parser)
    parser = cell_detect_parse(parser)
    parser = classification_parse(parser)
    parser = cube_extract_parse(parser)
    parser = misc_parse(parser)
    parser = figures_parse(parser)
    parser = download_parser(parser)

    # brainreg options
    parser = atlas_parse(parser)
    parser = geometry_parser(parser)
    parser = brainreg_backend_parse(parser)
    # This needs to be abstracted away into brainreg for multiple backends
    parser = niftyreg_parse(parser)

    return parser


def main_parse(parser):
    main_parser = parser.add_argument_group("General options")

    main_parser.add_argument(
        "-s",
        "--signal-planes-paths",
        dest="signal_planes_paths",
        type=str,
        nargs="+",
        required=True,
        help="Path to the directory of the signal files. Can also be a text"
        "file pointing to the files. For a 3d tiff, data is in z, y, x order",
    )
    main_parser.add_argument(
        "-b",
        "--background-planes-path",
        dest="background_planes_path",
        type=str,
        nargs=1,
        required=True,
        help="Path to the directory of the background files. Can also be a "
        "text file pointing to the files. For a 3d tiff, data is in z, y, x "
        "order",
    )
    main_parser.add_argument(
        "-o",
        "--output-dir",
        dest="output_dir",
        type=str,
        required=True,
        help="Output directory for all intermediate and final results.",
    )
    main_parser.add_argument(
        "--signal-channel-ids",
        dest="signal_ch_ids",
        type=check_positive_int,
        nargs="+",
        help="Channel ID numbers, in the same order as 'signal-planes-paths'."
        " Will default to '0, 1, 2' etc, but maybe useful to specify.",
    )
    main_parser.add_argument(
        "--background-channel-id",
        dest="background_ch_id",
        type=check_positive_int,
        help="Channel ID number, corresponding to 'background-planes-path'.",
    )
    main_parser.add_argument(
        "--version",
        action="version",
        version="%(prog)s {version}".format(version=__version__),
    )
    return parser


def pixel_parser(parser):
    # TODO: separate the two groups. Image pixel sizes are needed in lots of
    # places that the network pixel sizes are not
    pixel_opt_parser = parser.add_argument_group(
        "Options to define pixel sizes of raw data"
    )
    pixel_opt_parser.add_argument(
        "-v",
        "--voxel-sizes",
        dest="voxel_sizes",
        required=True,
        nargs=3,
        type=partial(check_positive_float, none_allowed=False),
        help="Voxel sizes in microns, in the order of data orientation "
        "(z, y, x). E.g. '5 2 2'",
    )
    pixel_opt_parser.add_argument(
        "--network-voxel-sizes",
        dest="network_voxel_sizes",
        nargs=3,
        type=partial(check_positive_float, none_allowed=False),
        default=[5, 1, 1],
        help="Voxel sizes in microns that the machine learning network was "
        "trained on, in the order of data orientation. e.g. '5 2 2'."
        "Set this to adjust the pixel sizes of the extracted cubes",
    )
    return parser


def run_parse(parser):
    run_parser = parser.add_argument_group(
        "Options to disable part of brainmapper"
    )
    run_parser.add_argument(
        "--no-detection",
        dest="no_detection",
        action="store_true",
        help="Dont run cell candidate detection",
    )

    run_parser.add_argument(
        "--no-classification",
        dest="no_classification",
        action="store_true",
        help="Do not run cell classification",
    )
    run_parser.add_argument(
        "--no-register",
        dest="no_register",
        action="store_true",
        help="Do not perform registration",
    )
    run_parser.add_argument(
        "--no-analyse",
        dest="no_analyse",
        action="store_true",
        help="Do not analyse and export cell positions",
    )
    run_parser.add_argument(
        "--no-figures",
        dest="no_figures",
        action="store_true",
        help="Do not create figures (e.g. heatmap)",
    )

    return parser


def io_parse(parser):
    io_parser = parser.add_argument_group("Input & output options")
    io_parser.add_argument(
        "--start-plane",
        dest="start_plane",
        type=check_positive_int,
        default=0,
        help="The first plane index to process in the Z dimension (inclusive, "
        "to process a subset of the data).",
    )
    io_parser.add_argument(
        "--end-plane",
        dest="end_plane",
        type=int,
        default=-1,
        help="The last plane to process in the Z dimension (exclusive, to "
        "process a subset of the data).",
    )
    return parser


def cell_detect_parse(parser):
    # TODO: improve the help on these files
    cell_detect_parser = parser.add_argument_group(
        "Detection options", description="Cell detection options"
    )
    cell_detect_parser.add_argument(
        "--save-planes",
        dest="save_planes",
        action="store_true",
        help="Whether to save the individual planes after "
        "processing and thresholding. Useful for debugging.",
    )

    cell_detect_parser.add_argument(
        "--outlier-keep",
        dest="outlier_keep",
        action="store_true",
        help="Dont remove putative cells that fall outside initial clusters",
    )

    cell_detect_parser.add_argument(
        "--artifact-keep",
        dest="artifact_keep",
        action="store_true",
        help="Save artifacts into the initial xml file",
    )

    cell_detect_parser.add_argument(
        "--max-cluster-size",
        dest="max_cluster_size",
        type=check_positive_int,
        default=100000,
        help="Largest detected cell cluster (in cubic um) where splitting "
        "should be attempted. Clusters above this size will be labeled as "
        "artifacts.",
    )

    cell_detect_parser.add_argument(
        "--soma-diameter",
        dest="soma_diameter",
        type=check_positive_float,
        default=16,
        help="The expected in-plane (xy) soma diameter (microns).",
    )

    cell_detect_parser.add_argument(
        "--ball-xy-size",
        dest="ball_xy_size",
        type=check_positive_int,
        default=6,
        help="3d filter's in-plane (xy) filter ball size (microns).",
    )
    cell_detect_parser.add_argument(
        "--ball-z-size",
        dest="ball_z_size",
        type=check_positive_int,
        default=15,
        help="3d filter's axial (z) filter ball size (microns).",
    )
    cell_detect_parser.add_argument(
        "--ball-overlap-fraction",
        dest="ball_overlap_fraction",
        type=check_positive_float,
        default=0.6,
        help="3d filter's fraction of the ball filter needed to be filled by "
        "foreground voxels, centered on a voxel, to retain the voxel.",
    )

    cell_detect_parser.add_argument(
        "--log-sigma-size",
        dest="log_sigma_size",
        type=check_positive_float,
        default=0.2,
        help="Gaussian filter width (as a fraction of soma diameter) used "
        "during 2d in-plane Laplacian of Gaussian filtering.",
    )
    cell_detect_parser.add_argument(
        "--threshold",
        dest="n_sds_above_mean_thresh",
        type=float,
        default=10,
        help="Per-plane intensity threshold (the number of standard "
        "deviations above the mean) of the filtered 2d planes used to mark "
        "pixels as foreground or background.",
    )
    cell_detect_parser.add_argument(
        "--tiled-threshold",
        dest="n_sds_above_mean_tiled_thresh",
        type=float,
        default=10,
        help="Per-plane, per-tile intensity threshold (the number of standard"
        " deviations above the mean) for the filtered 2d planes used to mark "
        "pixels as foreground or background. When used, (tile size is not "
        "zero) a pixel is marked as foreground if its intensity is above both"
        " the per-plane and per-tile threshold. I.e. it's above the set "
        "number of standard deviations of the per-plane average and of the "
        "per-plane per-tile average for the tile that contains it.",
    )
    cell_detect_parser.add_argument(
        "--tiled-threshold-tile-size",
        dest="tiled_thresh_tile_size",
        type=check_positive_float,
        default=None,
        help="The tile size used to tile the x, y plane to calculate the "
        "local average intensity for the tiled threshold. The value is "
        "multiplied by soma diameter (i.e. 1 means one soma diameter). If "
        "zero or None, the tiled threshold is disabled and only the per-plane"
        " threshold is used. Tiling is done with 50%% overlap when striding.",
    )
    cell_detect_parser.add_argument(
        "--soma-spread-factor",
        dest="soma_spread_factor",
        type=check_positive_float,
        default=1.4,
        help="Cell spread factor for determining the largest cell volume "
        "before splitting up cell clusters. Structures with spherical volume "
        "of diameter `soma_spread_factor * soma_diameter` or less will not "
        "be split.",
    )
    cell_detect_parser.add_argument(
        "--detection-batch-size",
        dest="detection_batch_size",
        type=check_positive_int,
        default=None,
        help="The number of planes of the original data volume to process at "
        "once. When None (the default), it defaults to 1 for GPU and 4 for "
        "CPU. The GPU/CPU memory must be able to contain this many planes "
        "for all the filters. For performance-critical applications, tune to "
        "maximize memory usage without running out. Check your GPU/CPU memory"
        " to verify it's not full.",
    )

    return parser


def classification_parse(parser):
    classification_parser = parser.add_argument_group(
        "Cell classification options"
    )
    classification_parser.add_argument(
        "--trained-model",
        dest="trained_model",
        type=str,
        help="Path to the trained model",
    )
    classification_parser.add_argument(
        "--model-weights",
        dest="model_weights",
        type=str,
        help="Path to existing model weights",
    )
    classification_parser.add_argument(
        "--network-depth",
        dest="network_depth",
        type=valid_model_depth,
        default="50",
        help="Resnet depth (based on He et al. (2015)",
    )
    classification_parser.add_argument(
        "--batch-size",
        dest="classification_batch_size",
        type=check_positive_int,
        default=64,
        help="Deprecated. Use classification-batch-size instead.",
    )
    classification_parser.add_argument(
        "--classification-batch-size",
        dest="classification_batch_size",
        type=check_positive_int,
        default=64,
        help="How many potential cells to classify at one time. The GPU/CPU "
        "memory must be able to contain at once this many data cubes for the "
        "models. For performance-critical applications, tune to maximize "
        "memory usage without running out. Check your GPU/CPU memory to "
        "verify it's not full.",
    )
    return parser


def cube_extract_parse(parser):
    cube_extract_parser = parser.add_argument_group("Cube extraction options")
    cube_extract_parser.add_argument(
        "--cube-width",
        dest="cube_width",
        type=check_positive_int,
        default=50,
        help="The width of the data cube centered on the cell, used for "
        "classification",
    )
    cube_extract_parser.add_argument(
        "--cube-height",
        dest="cube_height",
        type=check_positive_int,
        default=50,
        help="The height of the data cube centered on the cell, used "
        "for classification",
    )
    cube_extract_parser.add_argument(
        "--cube-depth",
        dest="cube_depth",
        type=check_positive_int,
        default=20,
        help="The depth of the data cube centered on the cell, used for "
        "classification",
    )
    cube_extract_parser.add_argument(
        "--save-empty-cubes",
        dest="save_empty_cubes",
        action="store_true",
        help="If a cube cannot be extracted (e.g. to close to the edge of"
        "the image), save an empty cube instead. Useful to keep track"
        "of all cell candidates.",
    )
    return parser


def config_parse(parser):
    config_opt_parser = parser.add_argument_group("Config options")
    config_opt_parser.add_argument(
        "--config",
        dest="registration_config",
        type=str,
        default=user_specific_configuration_path(),
        help="To supply your own, custom configuration file.",
    )

    return parser


def figures_parse(parser):
    figure_parser = parser.add_argument_group(
        "Figure generation specific parameters"
    )
    figure_parser.add_argument(
        "--heatmap-smoothing",
        dest="heatmap_smooth",
        type=check_positive_float,
        default=100,
        help="Gaussian smoothing sigma, in um.",
    )
    figure_parser.add_argument(
        "--no-mask-figs",
        dest="mask_figures",
        action="store_false",
        help="Don't mask the figures (removing any areas outside the brain,"
        "from e.g. smoothing)",
    )
    return parser


def misc_parse(parser):
    misc_parser = parser.add_argument_group("Misc options")
    misc_parser.add_argument(
        "--n-free-cpus",
        dest="n_free_cpus",
        type=check_positive_int,
        default=2,
        help="The number of CPU cores on the machine to leave "
        "unused by the program to spare resources.",
    )
    misc_parser.add_argument(
        "--torch-device",
        dest="torch_device",
        type=check_str,
        default=None,
        help="The device on which to run the computation. If not specified "
        "(None), cuda will be used if a GPU is available, otherwise cpu. You "
        "can also manually specify cuda or cpu.",
    )
    misc_parser.add_argument(
        "--pin-memory",
        dest="pin_memory",
        action="store_true",
        help="Pins data to be sent to the GPU to the CPU memory. This allows "
        "faster GPU data speeds, but can only be used if the data used by the"
        " GPU can stay in the CPU RAM while the GPU uses it. I.e. there's "
        "enough RAM. Otherwise, if there's a risk of the RAM being paged, it "
        "shouldn't be used. Defaults to False.",
    )
    misc_parser.add_argument(
        "--max-ram",
        dest="max_ram",
        type=check_positive_float,
        default=None,
        help="Maximum amount of RAM to use (in GB) - not currently fully "
        "implemented for all parts of brainmapper",
    )
    misc_parser.add_argument(
        "--save-csv",
        dest="save_csv",
        action="store_true",
        help="Save .csv files of cell locations (in addition to xml)."
        "Useful for importing into other software.",
    )
    misc_parser.add_argument(
        "--debug",
        dest="debug",
        action="store_true",
        help="Debug mode. Will increase verbosity of logging and save all "
        "intermediate files for diagnosis of software issues.",
    )
    misc_parser.add_argument(
        "--sort-input-file",
        dest="sort_input_file",
        action="store_true",
        help="If set to true, the input text file will be sorted using "
        "natural sorting. This means that the file paths will be "
        "sorted as would be expected by a human and "
        "not purely alphabetically",
    )
    return parser
